/*
 * Based on Obdev's AVRUSB code and under the same license.
 *
 * SpaceOrb.h is a library for Arduino that wraps around the AVRUSB driver from ObDev
 *
 * Adapted from the UsbKeyboard.h library from Phil Lindsay
 *
 * by Michel Gutlich
 *
 */
#ifndef ORB_DEVICE_H
#define ORB_DEVICE_H

#include <avr/pgmspace.h>

#include "usbdrv.h"
/* these added for compatibility with 0012 Arduino */
#include <avr/interrupt.h>
#include <string.h>
typedef uint8_t byte;
#include <util/delay.h>

#include "orb_buffer.h"
#include "HardwareSerial.h"

#define SIMULTANEOUS_KEYS 5
#define REPORT_ID_JOYSTICK 1
#define REPORT_ID_KEYBOARD 2
#define REPORT_ID_MOUSE 3
#define REPORT_SIZE_MOUSE 4
#define MAX_SINGLE_TRANSMISSION_SIZE 8
static uchar reportBuffer[10];				/* buffer for HID reports */
static uchar    idleRate;           // in 4 ms units 

#include "hid_keys.h"

PROGMEM char usbHidReportDescriptor[USB_CFG_HID_REPORT_DESCRIPTOR_LENGTH] = {  // USB report descriptor for a 6-axis, 4 button joystick
    0x05, 0x01,                    // USAGE_PAGE (Generic Desktop)
    0x09, 0x04,                    // USAGE (Joystick)
    0xa1, 0x01,                    // COLLECTION (Application)
    0x85, 0x01,                    // REPORT_ID (1)
    0xa1, 0x00,                    //   COLLECTION (Physical)
    0x09, 0x30,                    //     USAGE (X)
    0x09, 0x31,                    //     USAGE (Y)
    0x09, 0x32,                    //     USAGE (Z)
    0x09, 0x33,                    //     USAGE (Rx)
    0x09, 0x34,                    //     USAGE (Ry)
    0x09, 0x35,                    //     USAGE (Rz)
    0x15, 0x00,                    //     LOGICAL_MINIMUM (0)
    0x26, 0xff, 0x03,              //     LOGICAL_MAXIMUM (1023)
    0x75, 0x0a,                    //     REPORT_SIZE (10)
    0x95, 0x06,                    //     REPORT_COUNT (6)
    0x81, 0x02,                    //     INPUT (Data,Var,Abs)
    0xc0,                          //     END_COLLECTION
    0x05, 0x09,                    //   USAGE_PAGE (Button)
    0x19, 0x01,                    //   USAGE_MINIMUM (Button 1)
    0x29, 0x10,                    //   USAGE_MAXIMUM (Button 16)
    0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
    0x25, 0x01,                    //   LOGICAL_MAXIMUM (1)
    0x75, 0x01,                    //   REPORT_SIZE (1)
    0x95, 0x10,                    //   REPORT_COUNT (16)
    0x55, 0x00,                    //   UNIT_EXPONENT (0)
    0x65, 0x00,                    //   UNIT (None)
    0x81, 0x02,                    //   INPUT (Data,Var,Abs)
    0x75, 0x04,                    //   REPORT_SIZE (1)
    0x95, 0x01,                    //   REPORT_COUNT (1)
    0x81, 0x03,                    //   INPUT (Cnst,Var,Abs)
    0xc0,                           // END_COLLECTION

    // the keyboard report

    0x05, 0x01,                    // USAGE_PAGE (Generic Desktop) 
    0x09, 0x06,                    // USAGE (Keyboard) 
    0xa1, 0x01,                    // COLLECTION (Application) 
    0x85, 0x02,                    // REPORT_ID (2)
    0x05, 0x07,                    //   USAGE_PAGE (Keyboard) 
    0x19, 0xe0,                    //   USAGE_MINIMUM (Keyboard LeftControl) 
    0x29, 0xe7,                    //   USAGE_MAXIMUM (Keyboard Right GUI) 
    0x15, 0x00,                    //   LOGICAL_MINIMUM (0) 
    0x25, 0x01,                    //   LOGICAL_MAXIMUM (1) 
    0x75, 0x01,                    //   REPORT_SIZE (1) 
    0x95, 0x08,                    //   REPORT_COUNT (8) 
    0x81, 0x02,                    //   INPUT (Data,Var,Abs) 
    0x95, SIMULTANEOUS_KEYS,       //   REPORT_COUNT (simultaneous keystrokes) 
    0x75, 0x08,                    //   REPORT_SIZE (8) 
    0x25, 0x65,                    //   LOGICAL_MAXIMUM (101) 
    0x19, 0x00,                    //   USAGE_MINIMUM (Reserved (no event indicated)) 
    0x29, 0x65,                    //   USAGE_MAXIMUM (Keyboard Application) 
    0x81, 0x00,                    //   INPUT (Data,Ary,Abs) 
    0xc0,                           // END_COLLECTION 

    // the mouse report, 54 bytes
    0x05, 0x01,                    // USAGE_PAGE (Generic Desktop)
    0x09, 0x02,                    // USAGE (Mouse)
    0xa1, 0x01,                    // COLLECTION (Application)
    0x85, 0x03,                    // REPORT_ID (3)
    0x09, 0x01,                    //   USAGE (Pointer)
    0xA1, 0x00,                    //   COLLECTION (Physical)
    0x05, 0x09,                    //     USAGE_PAGE (Button)
    0x19, 0x01,                    //     USAGE_MINIMUM
    0x29, 0x03,                    //     USAGE_MAXIMUM
    0x15, 0x00,                    //     LOGICAL_MINIMUM (0)
    0x25, 0x01,                    //     LOGICAL_MAXIMUM (1)
    0x95, 0x03,                    //     REPORT_COUNT (3)
    0x75, 0x01,                    //     REPORT_SIZE (1)
    0x81, 0x02,                    //     INPUT (Data,Var,Abs)
    0x95, 0x01,                    //     REPORT_COUNT (1)
    0x75, 0x05,                    //     REPORT_SIZE (5)
    0x81, 0x03,                    //     INPUT (Const,Var,Abs)
    0x05, 0x01,                    //     USAGE_PAGE (Generic Desktop)
    0x09, 0x30,                    //     USAGE (X)
    0x09, 0x31,                    //     USAGE (Y)
    0x09, 0x38,                    //     USAGE (Wheel)
    0x15, 0x81,                    //     LOGICAL_MINIMUM (-127)
    0x25, 0x7F,                    //     LOGICAL_MAXIMUM (127)
    0x75, 0x08,                    //     REPORT_SIZE (8)
    0x95, 0x03,                    //     REPORT_COUNT (3)
    0x81, 0x06,                    //     INPUT (Data,Var,Rel)
    0xC0,                          //   END_COLLECTION
    0xC0                          // END COLLECTION

/* This is the same report descriptor as seen in a Logitech mouse. The data
 * described by this descriptor consists of 4 bytes:
 *      .  .  .  .  . B2 B1 B0 .... one byte with mouse button states  --> bm
 *     X7 X6 X5 X4 X3 X2 X1 X0 .... 8 bit signed relative coordinate x --> dx
 *     Y7 Y6 Y5 Y4 Y3 Y2 Y1 Y0 .... 8 bit signed relative coordinate y --> dy
 *     W7 W6 W5 W4 W3 W2 W1 W0 .... 8 bit signed relative coordinate wheel --> dw
 */

};

unsigned short reverse_word( unsigned short b );

PROGMEM const char spaceball_setup_string[] ="\rCB\rNT\rFT?\rFR?\rP@r@r\rMSSV\rZ\rBcCcCcC\r";


class Orb_device //: public HID_device 
{
 public:
  Orb_device () {
    
    usbInit();
      
    sei();

  }
    
  void safe_send_serial( const char* str )
    {
      for ( int i = 0; str[i] != 0; ++i )
	{
	  Serial.print( str[i], BYTE );
	}
    }

  //sends a PROGMEM string to the serial port
  void safe_send_serial_P( const char* str )
    {
      while ( pgm_read_byte( str ) != 0x00 )
	{
	  refresh();
	  Serial.print(pgm_read_byte(str++), BYTE);
	}
    }

  void refresh() {
    usbPoll();
  }

  virtual void send_joystick_report( unsigned short x,
				     unsigned short y,
				     unsigned short z,
				     unsigned short rx,
				     unsigned short ry,
				     unsigned short rz,
				     unsigned short btn )
    {
      while( !usbInterruptIsReady() )
	{
	  usbPoll();
	}
      unsigned char high;
      unsigned char low;
      unsigned char temp;
      //      unsigned short rbtn;
      reportBuffer[0] = REPORT_ID_JOYSTICK;

      high = x >> 8;
      low = x & 255;
      reportBuffer[1] = low;
      temp = high;

      high = y >> 6;
      low = y & 63;
      reportBuffer[2] = (low << 2) + temp;
      temp = high;

      high = z >> 4;
      low = y & 15;
      reportBuffer[3] = (low << 4) + temp;
      temp = high;
      
      high = rx >> 2;
      low = rx & 3;
      reportBuffer[4] = (low << 6) + temp;
      temp = high;

      high = 0;
      low = 0;
      reportBuffer[5] = temp;
      temp = high;

      high = ry >> 8;
      low = ry & 255;
      reportBuffer[6] = low+temp;
      temp = high;

      high = rz >> 6;
      low = rz & 63;
      reportBuffer[7] = (low << 2) + temp;
      temp = high;

      // now buttons.  This is tricky because logical_buttons
      // is an unsigned short from orb_buffer where button 0 is bit 0,
      // but HID wants button 0 to be the first bit sent
      //      rbtn = reverse_word(btn);
      reportBuffer[8] = (temp & 15 ) + ((btn & 15) << 4);
      reportBuffer[9] = (btn & 4080) >> 4; //(buttons1_8 & 15) + (buttons9_16 << 4);
      reportBuffer[10] = (btn >> 12);//0;//(rbtn & 15) << 4;//btn; //(buttons9_16&15);
      while (!usbInterruptIsReady())
	{
	  usbPoll();
	}
      if(usbInterruptIsReady()) {
	usbSetInterrupt(reportBuffer, MAX_SINGLE_TRANSMISSION_SIZE );
      }
      while (!usbInterruptIsReady())
	{
	  usbPoll();      
	}
      if (usbInterruptIsReady())
	{
	  usbSetInterrupt( reportBuffer+MAX_SINGLE_TRANSMISSION_SIZE, 3 );
	}
    }

  void send_keyboard_report( int count, byte* buttons )
    {
      if (count < MAX_SINGLE_TRANSMISSION_SIZE );
      memset( reportBuffer, 0, sizeof( reportBuffer ) );
      reportBuffer[0] = REPORT_ID_KEYBOARD;
      memcpy( reportBuffer+1, buttons, count );
      while (!usbInterruptIsReady())
	{
	  usbPoll();
	}
      usbSetInterrupt( reportBuffer, SIMULTANEOUS_KEYS + 2 );
    }

  void send_single_keyboard_report( byte keyStroke, byte modifiers)
    {
      
      while (!usbInterruptIsReady()) {
	// Note: We wait until we can send keystroke
	//       so we know the previous keystroke was
	//       sent.
	usbPoll();
      }
      
      memset(reportBuffer, 0, sizeof(reportBuffer));

      reportBuffer[0] = REPORT_ID_KEYBOARD;
      reportBuffer[1] = modifiers;
      reportBuffer[2] = keyStroke;
      reportBuffer[3] = 0;
      reportBuffer[4] = 0;
        
      usbSetInterrupt(reportBuffer, SIMULTANEOUS_KEYS+2);

      while (!usbInterruptIsReady()) {
	// Note: We wait until we can send keystroke
	//       so we know the previous keystroke was
	//       sent.
	usbPoll();
      }
      
      // This stops endlessly repeating keystrokes:
      memset(reportBuffer, 0, SIMULTANEOUS_KEYS+2 );
      reportBuffer[0] = REPORT_ID_KEYBOARD;
      usbSetInterrupt(reportBuffer, SIMULTANEOUS_KEYS+2);

    }

  void send_mouse_report( unsigned char buttons, 
			  char x, 
			  char y, 
			  char wheel )
    {
      while (!usbInterruptIsReady()) {
	// Note: We wait until we can send keystroke
	//       so we know the previous keystroke was
	//       sent.
	usbPoll();
      }
      memset( reportBuffer, 0, sizeof( reportBuffer ) );
      reportBuffer[0] = REPORT_ID_MOUSE;
      reportBuffer[1] = buttons;
      reportBuffer[2] = x;
      reportBuffer[3] = y;
      reportBuffer[4] = wheel;
      while (!usbInterruptIsReady())
	{
	  usbPoll();      
	}
      usbSetInterrupt(reportBuffer, REPORT_SIZE_MOUSE+1);

    }

  void sendRawPacket(byte val0, byte val1, byte val2, byte val3, byte val4 , byte val5 , byte val6, byte val7, byte val8, byte val9) {
      
    reportBuffer[0] = val0;
    reportBuffer[1] = val1;
    reportBuffer[2] = val2;
    reportBuffer[3] = val3;
    reportBuffer[4] = val4;
    reportBuffer[5] = val5;
    reportBuffer[6] = val6;
    reportBuffer[7] = val7;
    while (!usbInterruptIsReady())
      {
	usbPoll();
      }
    if(usbInterruptIsReady()) {
      usbSetInterrupt(reportBuffer, 8 );
    }
    while (!usbInterruptIsReady())
      {
	usbPoll();      
      }
    if (usbInterruptIsReady())
      {
	reportBuffer[0] = val8;
	reportBuffer[1] = val9;
	usbSetInterrupt( reportBuffer, 2 );
      }
  }
    
  //private: TODO: Make friend?
  uchar    reportBuffer[10];    // buffer for HID reports

};

Orb_device orb_device;

#ifdef __cplusplus
extern "C"{
#endif 
	
  // USB_PUBLIC uchar usbFunctionSetup
uchar usbFunctionSetup(uchar data[8]) 
  {
	usbRequest_t    *rq = (usbRequest_t *)((void *)data);
	  
    usbMsgPtr = orb_device.reportBuffer; //
    if((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS){ /* class request type */
      if(rq->bRequest == USBRQ_HID_GET_REPORT){ /* wValue: ReportType (highbyte), ReportID (lowbyte) */
		  reportBuffer[0]= 0;
		  reportBuffer[1]= 0;
		  reportBuffer[2]= 0;
		  reportBuffer[3]= 0;
		  reportBuffer[4]= 0;
		  reportBuffer[5]= 0;
		  reportBuffer[6]= 0;
		  reportBuffer[7]= 0;
	   
	return sizeof(reportBuffer);

      }else if(rq->bRequest == USBRQ_HID_GET_IDLE){
		usbMsgPtr = &idleRate;
		return 1;
      }else if(rq->bRequest == USBRQ_HID_SET_IDLE){
		idleRate = rq->wValue.bytes[1];
      }else{
		  /* no vendor specific requests implemented */
	}
    return 0;
  }
}
#ifdef __cplusplus
} // extern "C"
#endif


#endif // ORB_DEVICE_H
